import pytest
import numpy
from kaiju import RobotGrid, RobotGridCalib
from kaiju import utils
import time
numpy.random.seed(0)


def test_unevenCBs(plot=False):
    hasApogee = True
    greed = 0.8
    phobia = 0.2

    xPos, yPos = utils.hexFromDia(17, pitch=22.4)
    seed = 1
    # cb = 2.5
    cs = 0.04
    step = 0.1          # degrees per step in kaiju's rough path
    smoothPts = 10           # width of velocity smoothing window
    eps = step * 2.2
    # downsample = int(numpy.floor(50 / step))
    rg = RobotGrid(
        stepSize=step,
        epsilon=eps, seed=seed
    )

    for robotID, (x, y) in enumerate(zip(xPos, yPos)):
        rg.addRobot(robotID, str(robotID), [x, y, 0], hasApogee, collisionBuffer=numpy.random.uniform(1.5, 3.5))
    rg.initGrid()
    for rID in rg.robotDict:
        robot = rg.getRobot(rID)
        robot.setXYUniform()
    assert rg.getNCollisions() > 10

    rg.decollideGrid()

    for robot in rg.robotDict.values():
        robot.setDestinationAlphaBeta(0, 180)
    assert rg.getNCollisions() == 0
    tstart = time.time()
    rg.pathGenMDP(greed, phobia)
    print("pathgen took", time.time()-tstart)
    rg.smoothPaths(smoothPts)
    rg.simplifyPaths()
    rg.shrinkCollisionBuffer(cs)
    # rg.verifySmoothed()


    assert rg.smoothCollisions == 0
    print("n smooth collisions", rg.smoothCollisions)

    if plot:
        for r in rg.robotDict.values():
            utils.plotTraj(r, "unevenCBs", dpi=250)
        utils.plotPaths(rg, downsample=3, filename="unevenCBs.mp4")


if __name__ == "__main__":
    test_unevenCBs(plot=True)



import pytest
import numpy
import seaborn as sns
import matplotlib.pyplot as plt

from kaiju.robotGrid import RobotGridNominal
from kaiju import utils


def test_me(plot=False):
    angStep = 0.1
    downsample = int(numpy.floor(10 / angStep))
    seed = 4
    rg = RobotGridNominal(stepSize=angStep, seed=seed)
    rg.setCollisionBuffer(2)
    for r in rg.robotDict.values():
        r.setXYUniform()
        r.setDestinationAlphaBeta(0, 180)
    rg.decollideGrid()
    rg.pathGenMDP(0.8, 0.2)
    deadlockedRobots = []
    for r in rg.robotDict.values():
        # if not r.onTargetVec[-1]:
        if r.score() > 0:
            deadlockedRobots.append(r.id)

    if plot:
        utils.plotPaths(rg, downsample=downsample, filename="update.mp4")

        xs = []
        ys = []
        types = []
        for r in rg.robotDict.values():
            xs.append(r.basePos[0])
            ys.append(r.basePos[1])
            if r.hasApogee:
                types.append("BA")
            else:
                types.append("BOSS")

        for f in rg.fiducialDict.values():
            xs.append(f.xyzWok[0])
            ys.append(f.xyzWok[1])
            types.append("Fiducial")

        sns.scatterplot(x=xs, y=ys, hue=types)
        plt.savefig("scatterAPO.png")

    if len(deadlockedRobots):
        print("seed", seed, "failed with these", deadlockedRobots, "in ", rg.nSteps)
        assert False



if __name__ == "__main__":
    test_me(plot=True)



import pytest

from kaiju.robotGrid import RobotGrid
from kaiju import utils

import coordio


def test_uniqueFiducial():
    angStep = 1
    collisionBuffer = 2
    epsilon = angStep * 2
    hasApogee = True
    fiducialID = 1
    seed = 0
    rg = RobotGrid(angStep, epsilon, seed)
    rg.setCollisionBuffer(collisionBuffer)
    rg.addFiducial(fiducialID, [0, 0, 0])
    with pytest.raises(RuntimeError) as excinfo:
        rg.addFiducial(fiducialID, [30, 0, 0])
    assert "Fiducial ID already exists" in str(excinfo.value)

def test_fiducial(plot=False):
    angStep = 1
    collisionBuffer = 2
    fiducialCollisionBuffer = 1.5
    epsilon = angStep * 2
    hasApogee = True
    robotID = 1
    fiducialID = 10
    seed = 0
    rg = RobotGrid(angStep, epsilon, seed)
    rg.setCollisionBuffer(collisionBuffer)
    rg.addRobot(robotID, str(robotID), [0, 0, 0], hasApogee)
    rg.addFiducial(fiducialID, [22.4, 0, coordio.defaults.POSITIONER_HEIGHT], fiducialCollisionBuffer)
    rg.initGrid()
    robot = rg.getRobot(robotID)
    for betaAng in range(20):
        robot.setAlphaBeta(90,betaAng)
        rColliders = rg.robotColliders(robotID)
        fColliders = rg.fiducialColliders(robotID)
        if plot:
            utils.plotOne(0, rg, figname="fiducial_%i.png"%betaAng, isSequence=False, xlim=[-30, 30], ylim=[-30, 30])

        # assert len(rColliders) == 0
        # if betaAng < 14:
        #     assert fColliders == [fiducialID]
        # else:
        #     assert len(fColliders) == 0

def grow(plot=False):
    angStep = 1
    collisionBuffer = 2
    fiducialCollisionBuffer = 1.5
    epsilon = angStep * 2
    hasApogee = True
    robotID = 1
    fiducialID = 10
    seed = 0
    for cb in [1.5, 2, 2.5, 3]:
        rg = RobotGrid(angStep, epsilon, seed)
        rg.setCollisionBuffer(cb)
        rg.addRobot(robotID, str(robotID), [0, 0, 0], hasApogee)
        rg.addFiducial(fiducialID, [22.4, 0, coordio.defaults.POSITIONER_HEIGHT], fiducialCollisionBuffer)
        rg.initGrid()
        robot = rg.getRobot(robotID)
        robot.setAlphaBeta(90,0)
        if plot:
            utils.plotOne(0, rg, figname="grow_%.2f.png"%cb, isSequence=False, xlim=[-30, 30], ylim=[-30, 30])


if __name__ == "__main__":
    grow(plot=True)


import pytest

import kaiju
from kaiju.robotGrid import RobotGridNominal
from kaiju import utils


def test_collide(plot=False):
    # should make a grid
    rg = RobotGridNominal()
    collidedRobotIDs = []
    for rid, r in rg.robotDict.items():
      if r.holeID == "R-13C1":
        r.setAlphaBeta(90,0)
        collidedRobotIDs.append(rid)
      elif r.holeID == "R-13C2":
        collidedRobotIDs.append(rid)
        r.setAlphaBeta(270, 0)
      else:
        r.setAlphaBeta(90,180)

    for rid in collidedRobotIDs:
      assert rg.isCollided(rid)

    assert rg.getNCollisions() == 2

    if plot:
      utils.plotOne(0, rg, figname="test_collide.png", isSequence=False, xlim=[-30, 30], ylim=[-30, 30])

if __name__ == "__main__":
  test_collide(True)


import numpy

from kaiju import RobotGrid
from kaiju import utils



def test_rotGrid(plot=False):
    nDia = 15
    for ii, rotAng in enumerate(numpy.linspace(0,90,50)):
        xPos, yPos = utils.hexFromDia(nDia, pitch=22.4, rotAngle=rotAng)
        rg = RobotGrid()
        for robotID, (x, y) in enumerate(zip(xPos,yPos)):
            rg.addRobot(robotID, str(robotID), [x, y, 0], hasApogee=True)
        rg.initGrid()
        for robot in rg.robotDict.values():
            robot.setAlphaBeta(0,180)

        if plot:
            utils.plotOne(0, rg, figname="rotGrid_%s.png"%("%i"%ii).zfill(4), isSequence=False)

# def test_filledRotGrid(plot=False):
#     for ii, rotAng in enumerate(numpy.linspace(0,90,50)):
#         rg = utils.robotGridFromFilledHex(rotAngle=rotAng)
#         for robot in rg.robotDict.values():
#             robot.setAlphaBeta(0,180)
#         if plot:
#             utils.plotOne(0, rg, figname="rotFilledGrid_%s.png"%("%i"%ii).zfill(4), isSequence=False)


if __name__ == "__main__":
    test_rotGrid(plot=True)
    # test_filledRotGrid(plot=True)



#!/usr/bin/env python
# -*- coding:utf-8 -*-

# @Filename: robotGrid.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)


import os
import json
import pickle
import numpy as np
import fitsio
import matplotlib.pyplot as plt
import kaiju
import kaiju.cKaiju
from descartes import PolygonPatch
from shapely.geometry import LineString
import coordio
from coordio.defaults import IHAT, JHAT, KHAT, calibration
import pandas as pd
from scipy.signal import savgol_filter
from simplification.cutil import simplify_coords


# positionerTable = calibration.positionerTable
# wokCoords = calibration.wokCoords
# fiducialCoords = calibration.fiducialCoords

# __all__ = ['RobotGrid', 'RobotGridFilledHex']
# default orientation of positioner to wok
# iHat = [0,-1,0]
# jHat = [1,0,0]
# kHat = [0,0,1]

KAIJU_ETC_DIR = os.path.join(
    os.path.dirname(__file__), "etc"
)

# Create look-up dictionary for types to strings conversion
str2FiberType = {'ApogeeFiber': kaiju.cKaiju.ApogeeFiber,
                 'BossFiber': kaiju.cKaiju.BossFiber}
fiberType2Str = dict()
for k in str2FiberType.keys():
    fiberType2Str[str2FiberType[k]] = k


class RobotGrid(kaiju.cKaiju.RobotGrid):
    """Basic python subclass of cKaiju.RobotGrid

    Parameters:
    ----------

    stepSize : float, np.float32
        step size for paths (degrees), default 0.1.

    epsilon : float, or None
        distance metric for path simplification, smaller
        values yeild more points, default is stepsize * 2.2

    seed : int
        random seed for grid

    scaleFac : float
        multiplicative scale factor to model thermal expansion of wok,
        default is 1


    Attributes:
    ----------

    stepSize : float, np.float32
        step size for paths (degrees). A maximum perturbation allowed for either
        alpha or beta axes

    epsilon : float
        smoothing parameter used in the RDP path simplification.  Smaller values
        mean smaller deviations from the rough path are allowed.

    seed : int
        seed for random number generator when used

    robotDict : dictionary of Robot class objects
        all robots

    nRobots : int
        number of robots

    fiducialDict : dictionary of Fiducial objects
        positions of fiducials

    targetDict : dictionary of Target class objects
        targets in field, with ID as keys

    smoothCollisions : int
        number of collisions detected after attempted path smoothing, should
        be zero when things work out

    didFail : bool
        path generation failed, not all robots reached target

    nSteps : int
        steps taken to generate a path

    """
    def __init__(
        self, stepSize=0.1,
        epsilon=None, seed=0, scaleFac=1):

        np.random.seed(int(seed)) # for seedifying stuff out here in python

        self.stepSize = stepSize
        self.scaleFac = scaleFac
        if epsilon is None:
            self.epsilon = stepSize * 2.2
        else:
            self.epsilon = epsilon
        self.seed = seed
        self.totalReplaced = 0 # hack for now until it goes into the C++
        self.runtime = 0 # hack for now until
        super().__init__(self.stepSize, self.epsilon, self.seed)
        self._load_grid()
        return

    def _load_grid(self):
        # subclasses may want to specify grid initialization here
        # but they don't have to
        pass

    def getRandomPathPair(self, alphaHome=0, betaHome=180, betaLim=None):
        """
        Get a random trajectory pair.

        Parameters
        -------------
        alphaHome : float
            degrees for alpha axis lattice/home state
        betaHome : float
            degrees for beta axis lattice/home state
        betaLim : None, or [minBeta, maxBeta]  must be safe
            flag for making collisions impossible


        Returns
        ---------
        forwardPath: dict
            alpha/beta points for all robots, moving them from lattice state
            to target state (aligned with sources).  The out path.

        reversePath: dict
            alpha/beta points for all robots, moving them from target state
            to lattice state.  The back path.
        """
        # betaSafeLim = 165  # 155 should be safe even.
        for robot in self.robotDict.values():
            robot.setDestinationAlphaBeta(alphaHome, betaHome)
            if robot.isOffline:
                continue
            if betaLim is not None:
                alpha = np.random.uniform(0, 359.99)
                beta = np.random.uniform(betaLim[0], betaLim[1])
                robot.setAlphaBeta(alpha, beta)
            else:
                robot.setXYUniform()


        if betaLim is not None and self.getNCollisions() > 0:
            raise RuntimeError("betaLim specified, but collisions present")
        else:
            self.decollideGrid()
        return self.getPathPair()



    def getPathPair(
        self, speed=2, smoothPoints=5, collisionShrink=0.05, pathDelay=1,
        epsilon=None
    ):
        """
        Get paths in format that jaeger expects.  No checking is done, so
        whoever calls this should check things on the RobotGrid
        and decide what to do next.

        Parameters
        -----------
        speed: float
            RPM at output, how fast robots move, max speed is 3

        pathDely: float
            seconds, how far in the future to put the first point, this
            allows a robot to "catch up" to the expected starting point
            for the path, if it's not there already.

        Returns
        ---------
        toDestination: dict
            alpha/beta points in time for all robots.  Path begins at robot
            grid's initialized state, moving toward destination state.

        fromDestination: dict
            alpha/beta points in time for all robots.  Path begins at robot
            grid's destination state, moving toward the initialized state.
            This is just simply a reversed version of toDestination.
        """
        if epsilon is None:
            epsilon = self.epsilon

        toDestination = {}
        fromDestination = {}

        self.maxSpeed = speed * 360 / 60.
        _dt = self.stepSize / self.maxSpeed
        self.maxAccel = 2*self.maxSpeed / _dt

        for r in self.robotDict.values():
            # if robot is offline, don't get a path for it
            if r.isOffline:
                continue

            ap = [x[1] for x in r.alphaPath]
            bp = [x[1] for x in r.betaPath]
            # buffer ends
            ap = np.array([[ap[0]]*smoothPoints + ap + [ap[-1]]*smoothPoints]).flatten()
            bp = np.array([[bp[0]]*smoothPoints + bp + [bp[-1]]*smoothPoints]).flatten()
            steps = np.arange(len(ap))
            # smooth
            aps = savgol_filter(ap, smoothPoints, polyorder=3)
            bps = savgol_filter(bp, smoothPoints, polyorder=3)

            # simplify
            apss = simplify_coords(np.array([steps, aps]).T, epsilon)
            bpss = simplify_coords(np.array([steps, bps]).T, epsilon)
            # linearly interpolate back to original density
            # (for collision checking after smoothing/simplifying)
            apssi = np.interp(steps, apss[:,0], apss[:,1])
            bpssi = np.interp(steps, bpss[:,0], bpss[:,1])
            # set on the robot object for checking for collisions

            r.alphaPathBuffered = [[t,x] for t,x in zip(steps, ap)]
            r.betaPathBuffered = [[t,x] for t,x in zip(steps, bp)]

            r.smoothedAlphaPath = [[t,x] for t,x in zip(steps, aps)]
            r.smoothedBetaPath = [[t,x] for t,x in zip(steps, bps)]

            r.simplifiedAlphaPath = [[t,x] for t,x in apss]
            r.simplifiedBetaPath = [[t,x] for t,x in bpss]

            r.interpSimplifiedAlphaPath = [[t,x] for t,x in zip(steps, apssi)]
            r.interpSimplifiedBetaPath = [[t,x] for t,x in zip(steps, bpssi)]

            # calculate velocity and acceleration for each arm
            denseTimes = steps * self.stepSize / (speed * 360 / 60.)
            alphaVel = np.gradient(apssi)/_dt
            betaVel = np.gradient(bpssi)/_dt
            alphaAcc = np.gradient(alphaVel)/_dt
            betaAcc = np.gradient(betaVel)/_dt

            r.interpVelocityAlphaPath = [[t,x] for t,x in zip(denseTimes, alphaVel)]
            r.interpVelocityBetaPath = [[t,x] for t,x in zip(denseTimes, betaVel)]

            r.interpAccelerationAlphaPath = [[t,x] for t,x in zip(denseTimes, alphaAcc)]
            r.interpAccelerationBetaPath = [[t,x] for t,x in zip(denseTimes, betaAcc)]


            armPathToDest = {}
            armPathFromDest = {}

            for axis, data in zip(["alpha", "beta"], [apss, bpss]):
                # to destination path
                times = data[:, 0] * self.stepSize / (speed * 360 / 60.)
                angle = data[:, 1]
                armPathToDest[axis] = [(pos, time + pathDelay) for pos, time in zip(angle, times)]

                # from destination path, a reverse of the same thing
                timesR = np.abs(times - times[-1])[::-1]
                angleR = angle[::-1]
                armPathFromDest[axis] = [(pos, time + pathDelay) for pos, time in zip(angleR, timesR)]

            toDestination[r.id] = armPathToDest
            fromDestination[r.id] = armPathFromDest

        # self.shrinkCollisionBuffer(collisionShrink)
        # self.verifySmoothed(len(steps))
        # self.growCollisionBuffer(collisionShrink)

        return toDestination, fromDestination


        #     alphaPaths.append(ap)
        #     betaPaths.append(bp)

        # alphaPaths = np.array(alphaPaths)
        # betaPaths = np.array(betaPaths)
        # # smooth them
        # alphaSmooth = savgol_filter(alphaPaths, smoothPoints, polyorder=3, axis=1)
        # betaSmooth = savgol_filter(betaPaths, smoothPoints, polyorder=3, axis=1)

        # steps = numpy.arrange(alphaSmooth.shape[1])
        # # could potentially multiprocess this...
        # for ii, robotID in enumerate(robotIDs):
        #     _alphaSmooth = alphaSmooth[ii]
        #     _betaSmooth = betaSmooth[ii]
        #     alphaSimp =


    def _getPathPair(
        self, speed=2, smoothPoints=5, collisionShrink=0.05, pathDelay=1
    ):
        """
        Get paths in format that jaeger expects.  No checking is done, so
        whoever calls this should check things on the RobotGrid
        and decide what to do next.

        Parameters
        -----------
        speed: float
            RPM at output, how fast robots move, max speed is 3
        smoothPoints: int
            window width for smoothing a path's velocity profile,
            in units of steps.  Smooths out fast switching between forward
            and reverse moves on the axes.
        collisionShrink: float
            mm, how much to decrease the collision buffer to allow
            for smoothing
        pathDely: float
            seconds, how far in the future to put the first point, this
            allows a robot to "catch up" to the expected starting point
            for the path, if it's not there already.

        Returns
        ---------
        toDestination: dict
            alpha/beta points in time for all robots.  Path begins at robot
            grid's initialized state, moving toward destination state.

        fromDestination: dict
            alpha/beta points in time for all robots.  Path begins at robot
            grid's destination state, moving toward the initialized state.
            This is just simply a reversed version of toDestination.
        """
        # break out these parameters later
        # smoothPoints = 5
        # collisionShrink = 0.05  # mm, this value shoudld probably be a function of stepsize
        # speed = 2 # rpm at output
        # pathDelay = 1  # seconds
        ###########

        self.smoothPaths(smoothPoints)
        self.simplifyPaths()
        self.shrinkCollisionBuffer(collisionShrink)
        self.verifySmoothed(self.nSteps)
        # print("got %i smoothed collisions in getPathPair"%self.smoothCollisions)
        self.growCollisionBuffer(collisionShrink)

        toDestination = {}
        fromDestination = {}

        for r in self.robotDict.values():

            # if robot is offline, don't get a path for it
            if r.isOffline:
                continue

            ibp = np.array(r.simplifiedBetaPath)


            iap = np.array(r.simplifiedAlphaPath)


            alphaTimesR = iap[:, 0] * self.stepSize / (speed * 360 / 60.)
            alphaDegR = iap[:, 1]
            betaTimesR = ibp[:, 0] * self.stepSize / (speed * 360 / 60.)
            betaDegR = ibp[:, 1]


            # add time buffer for the reverse path, in case robot is
            # not exactly starting from the expected spot.
            # build path from initial state to destination
            armPathR = {}
            armPathR["alpha"] = [(pos, time + pathDelay) for pos, time in zip(alphaDegR, alphaTimesR)]
            armPathR["beta"] = [(pos, time + pathDelay) for pos, time in zip(betaDegR, betaTimesR)]

            toDestination[int(r.id)] = armPathR

            # build path from destination to initial state
            alphaTimesF = np.abs(alphaTimesR - alphaTimesR[-1])[::-1]
            alphaDegF = alphaDegR[::-1]
            betaTimesF = np.abs(betaTimesR - betaTimesR[-1])[::-1]
            betaDegF = betaDegR[::-1]

            armPathF = {}
            armPathF["alpha"] = [(pos, time + pathDelay) for pos, time in zip(alphaDegF, alphaTimesF)]
            armPathF["beta"] = [(pos, time + pathDelay) for pos, time in zip(betaDegF, betaTimesF)]


            fromDestination[int(r.id)] = armPathF

        return toDestination, fromDestination

    def addRobot(self,
        robotID, holeID, basePos, hasApogee=True,
        iHat=IHAT, jHat=JHAT, kHat=KHAT,
        dxyz=[0,0,0], alphaLen=coordio.defaults.ALPHA_LEN, alphaOffDeg=0,
        betaOffDeg=0, elementHeight=coordio.defaults.POSITIONER_HEIGHT,
        metBetaXY=coordio.defaults.MET_BETA_XY,
        bossBetaXY=coordio.defaults.BOSS_BETA_XY,
        apBetaXY=coordio.defaults.AP_BETA_XY,
        collisionSegBetaXY=None, collisionBuffer=2, lefthanded=False
    ):

        if collisionSegBetaXY is None:
            # find the point along the beta axis
            # that will "just" enclose the tip of the
            # beta arm + some "cornerProtection"
            tipFlat = 0.6  # mm, length of flat top beta arm
            betaPost = 0.5  # mm, radius of beta post that can catch fibers
            # how much space around tip to save, set to zero
            # to allow "corner brushes"?
            # cornerProtection = 0.3  # mm
            cornerProtection = 0.55
            dxTip = np.sqrt(collisionBuffer**2 + (tipFlat/2)**2) - cornerProtection
            #dxPost = r - betaPost
            dxPost = 0
            collisionSegBetaXY = [
                [dxPost, 0],  # beta shaft is 1mm diameter
                [coordio.defaults.BETA_LEN - dxTip, 0]
            ]

        else:
            # ensure it's listified
            collisionSegBetaXY = [list(x) for x in collisionSegBetaXY]

        return super().addRobot(
            robotID, holeID, list(basePos), list(iHat), list(jHat),
            list(kHat), list(dxyz), alphaLen, alphaOffDeg,
            betaOffDeg, elementHeight, self.scaleFac, list(metBetaXY),
            list(bossBetaXY), list(apBetaXY), collisionSegBetaXY,
            hasApogee, collisionBuffer, lefthanded
        )

    def addFiducial(self, fiducialID, xyzWok, collisionBuffer=1.5):
        return super().addFiducial(
            fiducialID, list(xyzWok), collisionBuffer
        )

    def plotRobotCore(self, ax=None, robotID=None, isCollided=False):
        """Basic plotting for a single robot

        Parameters:
        ----------

        ax : matplotlib axis object
            axis object to add robot to

        robotID : int, np.int32
            robotID of robot to plot

        isCollided : bool
            is this robot collided?

        Comments:
        --------

        Dark blue are assigned robots.
        Red are collided robots
        Light blue are unassigned robots
        """
        internalBuffer = 1.5

        robot = self.robotDict[robotID]
        alphaPoint = robot.betaCollisionSegment[0]
        betaPoint = robot.betaCollisionSegment[1]
        alphaX = alphaPoint[0]
        alphaY = alphaPoint[1]
        betaX = betaPoint[0]
        betaY = betaPoint[1]
        plt.plot([robot.xPos, alphaX], [robot.yPos, alphaY],
                 color='black', linewidth=2, alpha=0.5)

        topCollideLine = LineString([(alphaX, alphaY), (betaX, betaY)]
                                    ).buffer(internalBuffer, cap_style=1)
        topcolor = 'blue'
        edgecolor = 'black'
        if isCollided:
            topcolor = "red"
        if not robot.isAssigned():
            topcolor = "skyblue"
        patch = PolygonPatch(topCollideLine, fc=topcolor, ec=edgecolor,
                             alpha=0.5, zorder=10)
        ax.add_patch(patch)

        return

    def plotAllRobots(self):
        """Plots all the robots

        Comments:
        --------

        Dark blue are assigned robots.
        Red are collided robots
        Light blue are unassigned robots
        """
        plt.figure(figsize=(10, 10))
        ax = plt.gca()

        for robotID in self.robotDict.keys():
            isCollided = self.isCollided(robotID)
            self.plotRobotCore(ax=ax, robotID=robotID, isCollided=isCollided)

        rr = 340.
        plt.xlim(np.array([-1., 1.]) * rr)
        plt.ylim(np.array([-1., 1.]) * rr)
        return

    def tofits(self, filename=None):
        """Write robot information to FITS file

        Parameters:
        ----------

        filename : str
            FITS file name to write to (clobbers)

        Comments:
        --------

        HDU1 contains the robot information
        HDU2 contains the target information
        """
        robot_array = self.robot_array()
        target_array = self.target_array()
        fitsio.write(filename, robot_array, clobber=True)
        fitsio.write(filename, target_array, clobber=False)
        return

    def fromfits(self, filename=None):
        """Read robot information from FITS file

        Parameters:
        ----------

        filename : str
            FITS file name to read from

        Comments:
        --------

        Assumes the format written out by tofits()
        """
        robot_array = fitsio.read(filename, ext=1)
        target_array = fitsio.read(filename, ext=2)
        self.clearTargetList()
        self.target_fromarray(target_array)
        self.robot_fromarray(robot_array)
        return

    def singleRobotDict(self, robot, downsample=None):
        """Dictionary for a single robot

        Parameters:
        -----------
        robot : cKaiju.Robot
            robot instance to scrape

        downsample : int or None
            Down sample the paths by this factor to decrease
            data size, last step is always included

        Returns:
        -------------
        result : a dict

        Comments:
        ---------
        note: robot_dict is dictionary for all robots
        note: should get pybind to give this automatically
        using __dict__ but haven't figured it out yet
        note: need to figure out interp downsampling
        """
        if downsample is None:
            downsample = 1

        r = {}

        r["id"] = robot.id
        r["nDecollide"] = robot.nDecollide
        r["lastStepNum"] = robot.lastStepNum
        r["assignedTargetID"] = robot.assignedTargetID
        r["hasApogee"] = robot.hasApogee
        r["hasBoss"] = robot.hasBoss
        r["xPos"] = robot.xPos
        r["yPos"] = robot.yPos
        r["alpha"] = robot.alpha
        r["beta"] = robot.beta
        r["destinationAlpha"] = robot.destinationAlpha
        r["desstinationBeta"] = robot.destinationBeta
        r["angStep"] = robot.angStep
        r["collisionBuffer"] = robot.collisionBuffer
        r["alphaVel"] = robot.alphaVel
        r["betaVel"] = robot.betaVel

        # convert from numpy arrays to lists
        r["metFiberPos"] = list(robot.metFiberPos)
        r["bossFiberPos"] = list(robot.bossFiberPos)
        r["apFiberPos"] = list(robot.apFiberPos)
        # r["onTargetVec"] = robot.onTargetVec

        # sequences
        # convert from numpy arrays to lists
        downsampleMe = [
            "alphaPath",
            "betaPath",
            "roughAlphaX",
            "roughBetaX",
            "roughAlphaY",
            "roughBetaY"
        ]
        for item in downsampleMe:
            # these are all lists of numpy pairs
            l = [list(x) for x in getattr(robot, item)]
            lastElement = l[-1]
            l = l[::downsample] + [lastElement]
            r[item] = l

        # not sure how to handle downsampling for these yet
        r["smoothAlphaVel"] = robot.smoothAlphaVel
        r["smoothBetaVel"] = robot.smoothBetaVel
        # r["alphaPath"] = [list(x) for x in robot.alphaPath]
        # r["betaPath"] = [list(x) for x in robot.betaPath]
        r["smoothedAlphaPath"] = [list(x) for x in robot.smoothedAlphaPath]
        r["smoothedBetaPath"] = [list(x) for x in robot.smoothedBetaPath]
        r["simplifiedAlphaPath"] = [list(x) for x in robot.simplifiedAlphaPath]
        r["simplifiedBetaPath "] = [list(x) for x in robot.simplifiedBetaPath]
        r["interpSimplifiedAlphaPath"] = [list(x) for x in robot.interpSimplifiedAlphaPath]
        r["interpSimplifiedBetaPath "] = [list(x) for x in robot.interpSimplifiedBetaPath]
        r["interpAlphaX"] = [list(x) for x in robot.interpAlphaX]
        r["interpAlphaY"] = [list(x) for x in robot.interpAlphaY]
        r["interpBetaX"] = [list(x) for x in robot.interpBetaX]
        r["interpBetaY "] = [list(x) for x in robot.interpBetaY]
        # r["roughAlphaX"] = [list(x) for x in robot.roughAlphaX]
        # r["roughAlphaY"] = [list(x) for x in robot.roughAlphaY]
        # r["roughBetaX"] = [list(x) for x in robot.roughBetaX]
        # r["roughBetaY "] = [list(x) for x in robot.roughBetaY]
        r["interpCollisions "] = [list(x) for x in robot.interpCollisions]

        r["robotNeighbors"] = robot.robotNeighbors
        r["fiducialNeighbors"] = robot.fiducialNeighbors
        r["validTargetIDs"] = robot.validTargetIDs

        return r

    def robotGridSummaryDict(self):
        """Dictionary summary for this RobotGrid, contains
        less info and thus faster read/write than a full dict

        Returns:
        -------------
        result : a dict

        Comments:
        ---------
        note: should get pybind to give this automatically
        using __dict__ but haven't figured it out yet
        """
        rgd = self.robotGridDict(incRobotDict=False)
        robotDict = {}
        for rid, robot in self.robotDict.items():
            r = {}
            # print("rid", robot.id, robot.alpha, robot.beta)
            r["id"] = robot.id
            r["nDecollide"] = robot.nDecollide
            r["lastStepNum"] = robot.lastStepNum
            r["assignedTargetID"] = robot.assignedTargetID
            r["hasApogee"] = robot.hasApogee
            r["hasBoss"] = robot.hasBoss
            r["xPos"] = robot.xPos
            r["yPos"] = robot.yPos
            r["alpha"] = robot.alpha
            r["beta"] = robot.beta
            r["destinationAlpha"] = robot.destinationAlpha
            r["desstinationBeta"] = robot.destinationBeta
            r["angStep"] = robot.angStep
            r["collisionBuffer"] = robot.collisionBuffer

            # convert from numpy arrays to lists
            r["metFiberPos"] = list(robot.metFiberPos)
            r["bossFiberPos"] = list(robot.bossFiberPos)
            r["apFiberPos"] = list(robot.apFiberPos)
            r["roughAlphaX"] = robot.roughAlphaX[-1][-1]
            r["roughAlphaY"] = robot.roughAlphaY[-1][-1]
            r["roughBetaX"] = robot.roughBetaX[-1][-1]
            r["roughBetaY"] = robot.roughBetaY[-1][-1]
            # find the step at which this
            # guy found its target
            # otv = np.array(robot.onTargetVec)
            # print("otargvec", otv)
            # r["onTargetVec"] = robot.onTargetVec[-1]
            # # find last False (after which must be true)
            # # +1 because step numbers start from 1 and indices start from 0
            # whereFalse = np.argwhere(otv==False).flatten()
            # if not whereFalse:
            #     # empty list started on target? thats crazy
            #     lastFalse = 0
            # else:
            #     lastFalse = int(whereFalse[-1] + 1)
            # if lastFalse == len(otv):
            #     # last step was false robot never got there
            #     r["stepConverged"] = str(-1 )# -1 incicates never got to target
            # else:
            #     r["stepConverged"] = str(lastFalse)
            robotDict[robot.id] = r

        rgd["robotDict"] = robotDict
        return rgd

    def robotGridDict(self, downsample=None, incRobotDict=True):
        """Dictionary for this RobotGrid

        Parameters:
        ------------
        downsample : int or None
            Down sample the paths by this factor to decrease
            data size, last step is always included

        incRobotDict : Bool
            If false, don't include the robot dict

        Returns:
        -------------
        result : a dict

        Comments:
        ---------
        note: should get pybind to give this automatically
        using __dict__ but haven't figured it out yet
        """

        # get collision buffer from the first robot
        # no longer lives on the RobotGrid
        # note robots may have unique collision buffer sizes now...
        cb = list(self.robotDict.items())[0].collisionBuffer

        r = {}
        r["nRobots"] = self.nRobots
        r["epsilon"] = self.epsilon
        r["collisionBuffer"] = cb
        r["angStep"] = self.angStep
        r["didFail"] = self.didFail
        r["nSteps"] = self.nSteps
        r["maxPathSteps"] = self.maxPathSteps
        r["smoothCollisions"] = self.smoothCollisions
        r["maxDisplacement"] = self.maxDisplacement
        r["deadlockedRobots"] = self.deadlockedRobots()
        r["greed"] = self.greed
        r["phobia"] = self.phobia
        r["seed"] = self.seed
        r["algType"] = str(self.algType)
        r["totalReplaced"] = self.totalReplaced
        r["runtime"] = self.runtime
        if incRobotDict:
            robotDict = {}
            for rid, robot in self.robotDict.items():
                robotDict[rid] = self.singleRobotDict(robot, downsample)
            r["robotDict"] = robotDict
        return r

    def fullPickle(self, filename, downsample=None):
        """Pickle the output of self.robotGridDict

        Parameters:
        --------------
        filename : filename for the pickle

        downsample : int or None
            Down sample the paths by this factor to decrease
            data size, last step is always included

        """
        d = self.robotGridDict(downsample)
        with open(filename, "wb") as f:
            pickle.dump(d, f)

    def fullJSON(self, filename=None, downsample=None):
        """Format robotGridDict() output as json

        Parameters:
        ------------
        filename : string or None
            If not none, write json to file, otherwise
        downsample : int or None
            Down sample the paths by this factor to decrease
            data size, last step is always included

        Returns:
        -------------
        result : a json string or None

        """
        d = self.robotGridDict(downsample)
        if filename is not None:
            with open(filename, "w") as f:
                json.dump(d, f, separators=(',', ':'))
        else:
            return json.dumps(d)

    def summaryPickle(self, filename, downsample=None):
        """Pickle the output of self.robotGridDict

        Parameters:
        --------------
        filename : filename for the pickle

        downsample : int or None
            Down sample the paths by this factor to decrease
            data size, last step is always included

        """
        d = self.robotGridSummaryDict()
        with open(filename, "wb") as f:
            pickle.dump(d, f)

    def summaryJSON(self, filename=None):
        """Format robotGridDict() output as json

        Parameters:
        ------------
        filename : string or None
            If not none, write json to file, otherwise
        downsample : int or None
            Down sample the paths by this factor to decrease
            data size, last step is always included

        Returns:
        -------------
        result : a json string or None

        """
        d = self.robotGridSummaryDict()
        if filename is not None:
            with open(filename, "w") as f:
                json.dump(d, f, separators=(',', ':'))
        else:
            return json.dumps(d)

    def robot_dict(self):
        """Create dictionary with robot information

        Comments:
        --------

        Each value in dictionary contains a list with an element per robot.
        Fiducial information is not included.
        """
        robot_dict = dict()
        rd = self.robotDict
        ks = rd.keys()
        nks = len(ks)
        robot_dict['id'] = [int(0)] * nks
        robot_dict['xPos'] = [float(0)] * nks
        robot_dict['yPos'] = [float(0)] * nks
        robot_dict['alpha'] = [float(0)] * nks
        robot_dict['beta'] = [float(0)] * nks
        robot_dict['validTargetIDs'] = [[]] * nks
        robot_dict['assignedTargetID'] = [int(0)] * nks
        robot_dict['hasApogee'] = [int(0)] * nks
        robot_dict['hasBoss'] = [int(0)] * nks
        robot_dict['isAssigned'] = [int(0)] * nks
        robot_dict['assignedFiberType'] = ['none'] * nks
        robot_dict['isCollided'] = [int(0)] * nks
        for i, k in enumerate(ks):
            crd = rd[k]
            robot_dict['id'][i] = int(crd.id)
            robot_dict['xPos'][i] = float(crd.xPos)
            robot_dict['yPos'][i] = float(crd.yPos)
            robot_dict['alpha'][i] = float(crd.alpha)
            robot_dict['beta'][i] = float(crd.beta)
            robot_dict['validTargetIDs'][i] = [int(y) for y in crd.validTargetIDs]
            robot_dict['assignedTargetID'][i] = int(crd.assignedTargetID)
            robot_dict['hasApogee'][i] = int(crd.hasApogee)
            robot_dict['hasBoss'][i] = int(crd.hasBoss)
            robot_dict['isAssigned'][i] = int(crd.isAssigned())
            if(crd.assignedTargetID >= 0):
                robot_dict['assignedFiberType'][i] = fiberType2Str[self.targetDict[crd.assignedTargetID].fiberType]
            robot_dict['isCollided'][i] = int(self.isCollided(k))
        return(robot_dict)

    def target_dict(self):
        """Create dictionary with target information

        Comments:
        --------

        Each value in dictionary contains a list with an element per target.
        """
        ks = list(self.targetDict.keys())
        nks = len(ks)
        target_dict = dict()
        target_dict['id'] = [int(0)] * nks
        target_dict['x'] = [float(0)] * nks
        target_dict['y'] = [float(0)] * nks
        target_dict['fiberType'] = ['none'] * nks
        for i, k in enumerate(ks):
            ctd = self.targetDict[k]
            target_dict['id'][i] = int(ctd.id)
            target_dict['x'][i] = float(ctd.x)
            target_dict['y'][i] = float(ctd.y)
            target_dict['fiberType'][i] = fiberType2Str[ctd.fiberType]
        return(target_dict)

    def json(self):
        """Returns a javascript string setting robot and target dicts

        Comments:
        --------

        javascript string sets javascript dicts robot_str and target_obj

        Probably this should just return the jsons and not the "robot_obj ="
        and "target_obj =" parts.
        """
        robot_dict = self.robot_dict()
        target_dict = self.target_dict()
        json_str = '{'
        json_str = json_str + '"robot_obj" : ' + json.dumps(robot_dict) + ",\n"
        json_str = json_str + '"target_obj" : ' + json.dumps(target_dict) + "\n"
        json_str = json_str + '}'
        return(json_str)

    def html(self, filebase):
        """Write HTML format file for visualizing the current robot state

        Parameters:
        ----------

        filebase : str
            base of html and json to write to
        """
        fieldfile = filebase + '.json'
        fp = open(fieldfile, "w")
        fp.write(self.json())
        fp.close()

        # _htmlPath = os.path.join(
        #     os.path.dirname(__file__),
        #     "..",
        #     "..",
        #     "etc",
        #     "robotGrid.html"
        # )

        # fp = open(os.path.join(os.getenv('KAIJU_DIR'), 'etc',
        #                        'robotGrid.html'), "r")

        fp = open(os.path.join(KAIJU_ETC_DIR, "robotGrid.html"), "r")

        html_str = ''
        for l in fp.readlines():
            l = l.replace("fieldfile", "'" + os.path.basename(fieldfile) + "'")
            html_str = html_str + l
        fp.close()

        fp = open(filebase + '.html', "w")
        fp.write(html_str)
        fp.close()

        fp = open(os.path.join(KAIJU_ETC_DIR, 'robotGrid.js'), "r")
        js_str = ''
        for l in fp.readlines():
            js_str = js_str + l
        fp.close()

        fp = open(os.path.join(os.path.dirname(filebase), 'robotGrid.js'), "w")
        fp.write(js_str)
        fp.close()
        return

    def robot_array(self):
        """Return ndarray with robot and assignment information"""
        robot_dtype = np.dtype([('robotID', np.int32),
                                ('xPos', np.float32),
                                ('yPos', np.float32),
                                ('hasApogee', np.bool),
                                ('hasBoss', np.bool),
                                ('assignedTargetID', np.int64),
                                ('isAssigned', np.bool),
                                ('isCollided', np.bool),
                                ('alpha', np.float32),
                                ('beta', np.float32)])

        robot_array = np.zeros(self.nRobots, dtype=robot_dtype)
        ks = self.robotDict.keys()
        for i, k in enumerate(ks):
            crd = self.robotDict[k]
            robot_array[i]['robotID'] = crd.id
            robot_array[i]['xPos'] = crd.basePos[0]
            robot_array[i]['yPos'] = crd.basePos[1]
            robot_array[i]['hasApogee'] = crd.hasApogee
            robot_array[i]['hasBoss'] = crd.hasBoss
            robot_array[i]['assignedTargetID'] = crd.assignedTargetID
            robot_array[i]['isAssigned'] = crd.isAssigned()
            robot_array[i]['isCollided'] = self.isCollided(crd.id)
            robot_array[i]['alpha'] = crd.alpha
            robot_array[i]['beta'] = crd.beta
        return(robot_array)

    def target_array(self):
        """Return ndarray with target information"""
        target_dtype = np.dtype([('targetID', np.int64),
                                 ('x', np.float64),
                                 ('y', np.float64),
                                 ('z', np.float64),
                                 ('priority', np.int32),
                                 ('fiberType', np.string_, 30)])

        target_array = np.zeros(len(self.targetDict), dtype=target_dtype)
        i = 0
        for k, td in self.targetDict.items():
            td = self.targetDict[k]
            target_array['targetID'][i] = td.id
            target_array['x'][i] = td.xWok
            target_array['y'][i] = td.yWok
            target_array['z'][i] = td.zWok
            target_array['priority'][i] = td.priority
            ft = fiberType2Str[td.fiberType]
            target_array['fiberType'][i] = ft
            i = i + 1
        return(target_array)

    def target_fromarray(self, target_array):
        """Fill robotGrid targetDict from target array

        Parameters:
        ----------

        target_array : ndarray
            array with target information

        Comments:
        --------

        Does not clearTargetList() first, so will add targets to
        existing list.
        """
        for it in np.arange(len(target_array)):
            ft = target_array['fiberType'][it].decode().strip()
            fiberType = str2FiberType[ft]
            xyzWok = [
                target_array['x'][it],
                target_array['y'][it],
                target_array['z'][it]
            ]
            self.addTarget(targetID=target_array['targetID'][it],
                           xyzWok=xyzWok,
                           fiberType=fiberType,
                           priority=target_array['priority'][it])
        return

    def robot_fromarray(self, robot_array):
        """Assign robots according to robot_array

        Parameters:
        ----------

        robot_array : ndarray
            array with robot information

        Comments:
        --------

        Assumes that the robotRict is already created, and is
        consistent with the robot_array information. It double
        checks that any information in robot_array is consistent
        with this robotGrid.
        """
        for ir, r in enumerate(robot_array):
            if(r['isAssigned']):
                indx = r['assignedTargetID']
                self.assignRobot2Target(r['robotID'], indx)
            else:
                self.robotDict[r['robotID']].setAlphaBeta(r['alpha'],
                                                          r['beta'])
        r = self.robot_array()
        for n in r.dtype.names:
            if(n in robot_array.dtype.names):
                ibad = np.where(r[n] != robot_array[n])[0]
                if(len(ibad) > 0):
                    raise RuntimeError("Inconsistency in robot file in column {n}".format(n=n))
        return

    def _load_from_tables(self, positionerTable, wokCoords, fiducialCoords):

        df = pd.merge(positionerTable, wokCoords, on="holeID").reset_index()

        for ii, row in df.iterrows():
            self.addRobot(
                robotID=row.positionerID,
                holeID=row.holeID,
                basePos=[row.xWok, row.yWok, row.zWok],
                hasApogee=bool("Apogee" in row.holeType),
                iHat=[row.ix, row.iy, row.iz],
                jHat=[row.jx, row.jy, row.jz],
                kHat=[row.kx, row.ky, row.kz],
                dxyz=[row.dx, row.dy, 0],
                alphaLen=row.alphaArmLen,
                alphaOffDeg=row.alphaOffset,
                betaOffDeg=row.betaOffset,
                elementHeight=coordio.defaults.POSITIONER_HEIGHT,
                metBetaXY=[row.metX, row.metY],
                bossBetaXY=[row.bossX, row.bossY],
                apBetaXY=[row.apX, row.apY],
                collisionSegBetaXY=None,
                collisionBuffer=2
            )

        for ii, row in fiducialCoords.iterrows():
            self.addFiducial(
                fiducialID=int(row.id.strip("F")),
                xyzWok=[row.xWok, row.yWok, row.zWok],
                collisionBuffer=3
            )

        # gfa locations hard-coded for now
        # eventually read from a table
        gfaList = [
            [
                [-267.49, 89.15, coordio.defaults.POSITIONER_HEIGHT],
                [-210.95, 187.08, coordio.defaults.POSITIONER_HEIGHT]
            ],
            [
                [-56.54, 276.23, coordio.defaults.POSITIONER_HEIGHT],
                [56.54, 276.23, coordio.defaults.POSITIONER_HEIGHT]
            ],
            [
                [210.95, 187.07, coordio.defaults.POSITIONER_HEIGHT],
                [267.49, 89.15, coordio.defaults.POSITIONER_HEIGHT]
            ],
            [
                [267.49, -89.15, coordio.defaults.POSITIONER_HEIGHT],
                [210.95, -187.08, coordio.defaults.POSITIONER_HEIGHT]
            ],
            [
                [56.54, -276.23, coordio.defaults.POSITIONER_HEIGHT],
                [-56.54, -276.23, coordio.defaults.POSITIONER_HEIGHT]
            ],
            [
                [-210.95, -187.07, coordio.defaults.POSITIONER_HEIGHT],
                [-267.49, -89.15, coordio.defaults.POSITIONER_HEIGHT]
            ]
        ]

        for ii, collisionSegXYZWok in enumerate(gfaList):
            self.addGFA(ii, collisionSegXYZWok, collisionBuffer=3)

        # loop through grid and find special robots with a fiducial
        # located down and right that need special attention for
        # path planning / allowed placement.  These robots have a small
        # range of alpha/beta that are blocked due to robot range of
        # motion
        fcRows = fiducialCoords.row.to_numpy()
        fcCols = fiducialCoords.col.to_numpy()
        posIDs = []
        for fcRow, fcCol in zip(fcRows,fcCols):
            if np.isnan(fcRow) or np.isnan(fcCol):
                continue
            pRow = int(fcRow+1)
            # not sure why, but positive rows have columns
            # offset by 1 with respect to negative rows
            # probably a bug in the config file...
            if pRow > 0:
                pCol = int(fcCol-1)
            else:
                pCol = int(fcCol)
            sign = ""
            if pRow > 0:
                sign = "+"
            posIDs.append("R%s%iC%i"%(sign,pRow,pCol))

        for r in self.robotDict.values():
            if r.holeID in posIDs:
                r.fiducialWatch = True

        self.initGrid()

    def plot_state(
            self, highlightRobot=None, plotRobotIDs=True,
            returnax=True, figname="kaijuGrid.pdf"
        ):
        """Plot the current state of the robot grid

        Unassigned robots are skyblue
        Robots at their destination are gold
        Robots that are kaiju-collided are red
        Offline robots are black

        Parameters
        ------------
        highlightRobot : int, list, or None
            Robot id to highlight in Orange

        plotRobotIDs : bool
            If True, plot the robotIDs

        returnax : bool
            If True, return the matplotlib axis (no figure is written to disk)

        figname : str
            If returnax is False, write the image to disk with this path.
            Defaults to kaijuGrid.png.

        Returns
        ---------
        ax : matplotlib axes instance if returnax is True

        """
        from kaiju.utils import plotOne
        if returnax:
            ax = plotOne(
                step=0, robotGrid=self, isSequence=False,
                plotRobotIDs=plotRobotIDs,
                highlightRobot=highlightRobot, returnax=True
            )
            return ax
        else:
            plotOne(
                step=0, robotGrid=self, figname=figname, isSequence=False,
                plotRobotIDs=plotRobotIDs,
                highlightRobot=highlightRobot
            )


class RobotGridCalib(RobotGrid):

    def _load_grid(self):
        self._load_from_tables(
            calibration.positionerTable,
            calibration.wokCoords,
            calibration.fiducialCoords
        )


class RobotGridAPO(RobotGrid):

    def _load_grid(self):

        self._load_from_tables(
            calibration.positionerTable.loc["APO"],
            calibration.wokCoords.loc["APO"],
            calibration.fiducialCoords.loc["APO"]
        )


class RobotGridLCO(RobotGrid):

    def _load_grid(self):
        self._load_from_tables(
            calibration.positionerTable.loc["LCO"],
            calibration.wokCoords.loc["LCO"],
            calibration.fiducialCoords.loc["LCO"]
        )


class RobotGridNominal(RobotGrid):
    # constructed from a "perfect" grid
    # useful for testing because calib grid will change overtime
    # this is a flat wok model, with GFAs
    def _load_grid(self):

        positionerTable = pd.read_csv(
            os.path.join(KAIJU_ETC_DIR, "positionerTable.csv")
        )

        wokCoords = pd.read_csv(
            os.path.join(KAIJU_ETC_DIR, "wokCoords.csv")
        )

        fiducialCoords = pd.read_csv(
            os.path.join(KAIJU_ETC_DIR, "fiducialCoords.csv")
        )

        self._load_from_tables(positionerTable, wokCoords, fiducialCoords)






from kaiju.robotGrid import RobotGridNominal
from kaiju import utils
import psutil
import os
import time


nDia = 15
angStep = 1
collisionBuffer = 2
epsilon = angStep * 2
hasApogee = True
GBperByte = 1e-9 / 5

nGrids = int(300 / 5)

def test_memory():
    # pick a big grid, run a bunch of grids
    # watch memory...
    # xPos, yPos = utils.hexFromDia(19, pitch=22.4)
    for seed in range(nGrids):
        rg = RobotGridNominal(angStep, seed)
        rg.setCollisionBuffer(collisionBuffer)
        # for robotID, (x, y) in enumerate(zip(xPos, yPos)):
        #     rg.addRobot(robotID, str(robotID), [x, y,  hasApogee)
        #     # rg.robotDict[robotID].setTargetAlphaBeta(0, 180)
        # rg.initGrid()

        for r in rg.robotDict.values():
            r.setXYUniform()
            r.setDestinationAlphaBeta(10,170)
        rg.decollideGrid()
        rg.pathGenGreedy()
    process = psutil.Process(os.getpid())
    GBUsed = process.memory_info().rss * GBperByte
    # print("GBUSED", GBUsed)
    assert GBUsed < 0.3


if __name__ == "__main__":
    tstart = time.time()
    test_memory()
    print("took %.2f mins"%((time.time()-tstart)/60.))


import pytest
import coordio

from kaiju.robotGrid import RobotGrid, RobotGridNominal
from kaiju import utils

nDia = 15
angStep = 1
collisionBuffer = 2
epsilon = angStep * 2
seed = 0
hasApogee = True


def test_hexGrid():
    xPos, yPos = utils.hexFromDia(15, pitch = 22.4)
    rg = RobotGrid(angStep, epsilon, seed)
    for robotID, (x, y) in enumerate(zip(xPos, yPos)):
        basePos = [x, y, 0]
        rg.addRobot(
            robotID, str(robotID), basePos, hasApogee,
            collisionBuffer=collisionBuffer
        )
    rg.initGrid()


def test_doubleRobotID():
    # should raise runtime error when two robots have same id
    rg = RobotGrid(angStep, epsilon, seed)
    robotID = 1
    rg.addRobot(
        robotID, str(robotID), [0, 0, 0], hasApogee,
        collisionBuffer=collisionBuffer
    )
    with pytest.raises(RuntimeError) as excinfo:
        rg.addRobot(robotID, str(robotID), [30, 0, 0], hasApogee)
    assert "Robot ID already exists" in str(excinfo.value)

def test_filledHexGrid():
    apo = RobotGridNominal(angStep)

def test_doubleFiducialID():
    # should raise runtime error when two fiducials have same id
    rg = RobotGrid(angStep, epsilon, seed)
    fiducialID = 1
    rg.addFiducial(fiducialID, [0,0,coordio.defaults.POSITIONER_HEIGHT])
    with pytest.raises(RuntimeError) as excinfo:
        rg.addFiducial(fiducialID, [100,0,coordio.defaults.POSITIONER_HEIGHT])
    assert "Fiducial ID already exists" in str(excinfo.value)

if __name__ == "__main__":
    test_doubleRobotID()


import pytest

import numpy
import matplotlib.pyplot as plt

from kaiju.cKaiju import BossFiber, ApogeeFiber, MetrologyFiber
from kaiju.robotGrid import RobotGrid, RobotGridNominal
from kaiju import utils

import coordio

fpZ = coordio.defaults.POSITIONER_HEIGHT

def test_nonInit():
    print("test nonInit")
    rg = RobotGrid()
    rxyz = [0,0,0]
    rg.addRobot(robotID=1, holeID="none", basePos=rxyz)
    fxyz = [22.4, 0, fpZ]
    rg.addFiducial(fiducialID=1, xyzWok=fxyz)
    with pytest.raises(RuntimeError) as excinfo:
        txyz = [0,10,fpZ]
        rg.addTarget(targetID=1, xyzWok=txyz, fiberType=BossFiber, priority=1)
    assert "Initialize RobotGrid before adding targets" in str(excinfo.value)


def test_doubleTargetID():
    print("test doubleTargetID")
    rg = RobotGrid()
    rg.addRobot(robotID=1, holeID="none", basePos=[0,0,0])
    rg.addFiducial(fiducialID=1, xyzWok=[22.4,0,fpZ])
    rg.initGrid()
    rg.addTarget(targetID=1, xyzWok=[0, 10, fpZ], fiberType=BossFiber, priority=1)
    with pytest.raises(RuntimeError) as excinfo:
        rg.addTarget(targetID=1, xyzWok=[0, 10, fpZ], fiberType=BossFiber, priority=1)
    assert "Target ID already exists" in str(excinfo.value)


def test_addTarget():
    print("test addTarget")
    rg = RobotGrid()
    rg.addRobot(robotID=1, holeID="none", basePos=[0,0,0])
    rg.addFiducial(fiducialID=1, xyzWok=[22.4, 0, fpZ])
    rg.initGrid()
    rg.addTarget(targetID=1, xyzWok=[0, 10, fpZ], fiberType=BossFiber, priority=1)
    rg.addTarget(2, [10, 10, fpZ], ApogeeFiber)
    rg.addTarget(3, [-20, -40, fpZ], MetrologyFiber)
    assert True

def generateTargs(nTargs, fiberType):
    rg = RobotGridNominal()
    # put in range -300 300
    randomX = numpy.random.random_sample(nTargs)*650 - 325
    randomY = numpy.random.random_sample(nTargs)*560 - 280
    for tid, (x, y) in enumerate(zip(randomX, randomY)):
        rg.addTarget(tid, [x, y, fpZ], fiberType)
    return rg

def plotTargs(rg, figname):
    unreachableTargets = rg.unreachableTargets()
    xyGood = []
    xyBad = []
    for tid, target in rg.targetDict.items():
        if tid in unreachableTargets:
            xyBad.append([target.xWok, target.yWok])
        else:
            xyGood.append([target.xWok, target.yWok])
    xyGood = numpy.array(xyGood)
    xyBad = numpy.array(xyBad)
    plt.figure(figsize=(8,8))
    plt.plot(xyGood[:,0], xyGood[:,1], 'og')
    plt.plot(xyBad[:,0], xyBad[:,1], 'oc')
    # overplot fiducials
    for fid, fiducial in rg.fiducialDict.items():
        plt.plot(fiducial.xWok, fiducial.yWok, 'kx')
    # plt.xlim([-300, 300])
    # plt.ylim([-300, 300])
    ax = plt.gca()
    ax.set_aspect("equal")
    plt.savefig(figname, dpi=250)


def test_tonsOBosstargs(plot=False):
    print("test tonsOboss")
    nTargs = 100000
    rg = generateTargs(nTargs, BossFiber)
    unreachableTargets = rg.unreachableTargets()
    unFrac = len(unreachableTargets)/nTargs
    assert numpy.abs(unFrac - 0.29457) < 0.01
    if plot:
        plotTargs(rg, "tonOBoss.png")


def test_tonsOAptargs(plot=False):
    print("test tonsOAp")
    nTargs = 100000
    rg = generateTargs(nTargs, ApogeeFiber)
    unreachableTargets = rg.unreachableTargets()
    unFrac = len(unreachableTargets)/nTargs
    assert numpy.abs(unFrac - 0.31379) < 0.01
    if plot:
        plotTargs(rg, "tonOAp.png")


def test_validRobots(plot=False):
    print("test valid robots")
    nTargs = 50000
    rg = generateTargs(nTargs, BossFiber)
    assert len(rg.targetlessRobots()) == 0
    for rid, robot in rg.robotDict.items():
        assert len(robot.validTargetIDs) > 100
    if plot:
        # show a set of targets for a subset of robots
        rIDs = list(rg.robotDict.keys())
        for ii in [20, 120, 450]:
            robot = rg.robotDict[rIDs[ii]]
            targetXYs = []
            for tid in robot.validTargetIDs:
                target = rg.targetDict[tid]
                targetXYs.append([target.xWok, target.yWok])
            targetXYs = numpy.array(targetXYs)
            plt.plot(targetXYs[:,0], targetXYs[:,1], 'og', markersize=1, alpha=0.5)
        ax = plt.gca()
        plt.xlim([-350, 350])
        ax.set_aspect("equal")
        plt.savefig("subsetTarg.png", dpi=250)

def test_targetAssign(plot=False):
    # rg = utils.robotGridFromFilledHex()
    print("test target assign")
    rg = RobotGridNominal()
    for robot in rg.robotDict.values():
        robot.setXYUniform()
    rg.decollideGrid()
    if plot:
        utils.plotOne(0, rg, figname="beforeAssign.png", isSequence=False)#, highlightRobot=205)
    # find boss target positions for all robots
    rids = []
    for rID, robot in rg.robotDict.items():
        rids.append(rID)
        xyzTarg = robot.bossWokXYZ
        # give the target the same id as the robot
        rg.addTarget(rID, xyzTarg, BossFiber)
        targ = rg.targetDict[rID]
        assert rID in targ.validRobotIDs
        assert rID in robot.validTargetIDs
        rg.assignRobot2Target(robotID=rID, targID=targ.id)

    unreachableTargets = rg.unreachableTargets()
    assert len(unreachableTargets) == 0
    targetlessRobots = rg.targetlessRobots()
    assert len(targetlessRobots) == 0

    # unassign target 100
    rg.unassignTarget(rids[0])
    assert rg.robotDict[rids[0]].isAssigned() == False
    assert rg.targetDict[rids[0]].isAssigned() == False

    rg.unassignTarget(rids[1])
    assert rg.robotDict[rids[1]].isAssigned() == False
    assert rg.targetDict[rids[1]].isAssigned() == False

    assignedTargets = rg.assignedTargets()
    assert (rids[1] in assignedTargets) == False
    assert (rids[0] in assignedTargets) == False

    if plot:
        utils.plotOne(0, rg, figname="afterAssign.png", isSequence=False)

"""
robot alpha beta (138.53, 1.66) xpos/ypos (-33.6 -252.1866) number: 205 valid targs []
alphabeta [138.5303302264647, 1.6620349800302785] boss fiber not in valid list?
i think issue is it is at nearly full extension?

test targets at full extension?
"""
if __name__ == "__main__":
    # test_validRobots(plot=True)
    test_tonsOBosstargs(plot=True)
    # test_tonsOAptargs(plot=True)
    # test_nonInit()
    # test_doubleTargetID()
    # test_tonsOBosstargs(plot=True)
    # test_tonsOAptargs(plot=True)
    # test_targetAssign(plot=True)



import pytest
import numpy
import time

from kaiju.robotGrid import RobotGrid, RobotGridNominal
from kaiju import utils

# nDia = 15
# angStep = 1
# collisionBuffer = 1.8 #2
# epsilon = angStep * 2
hasApogee = True


def test_hexDeadlockedPath(plot=False):
    xPos, yPos = utils.hexFromDia(35, pitch=22.4)
    seed = 3
    cb = 2.5
    angStep = 1
    downsample = int(numpy.floor(3 / angStep))
    rg = RobotGrid(angStep, seed=seed)
    for robotID, (x, y) in enumerate(zip(xPos, yPos)):
        rg.addRobot(robotID, str(robotID), [x,y,0], hasApogee)
        rg.robotDict[robotID].setDestinationAlphaBeta(0, 180)
    rg.setCollisionBuffer(cb)
    rg.initGrid()
    for rID in rg.robotDict:
        robot = rg.getRobot(rID)
        robot.setXYUniform()
    assert rg.getNCollisions() > 10
    rg.decollideGrid()
    assert rg.getNCollisions() == 0
    rg.pathGenGreedy()
    assert rg.didFail
    if plot:
        utils.plotPaths(rg, downsample=downsample, filename="test_hexDeadlockedPath.mp4")
    # print("rg did fail", rg.didFail)


def test_pathGen(plot=False):
    xPos, yPos = utils.hexFromDia(15, pitch=22.4)
    print("using", len(xPos), "robots")
    seed = 3
    smoothPts = 5
    collisionShrink = 0.03
    angStep = 0.1
    collisionBuffer = 2
    # epsilon = angStep * 2
    downsample = int(numpy.floor(3 / angStep))
    rg = RobotGrid(angStep, seed=seed)
    for robotID, (x, y) in enumerate(zip(xPos, yPos)):
        rg.addRobot(robotID, str(robotID), [x, y, 0], hasApogee)
        rg.robotDict[robotID].setDestinationAlphaBeta(0, 180)
    rg.setCollisionBuffer(collisionBuffer)
    rg.initGrid()
    for rID in rg.robotDict:
        robot = rg.getRobot(rID)
        robot.setXYUniform()
    assert rg.getNCollisions() > 10
    if plot:
        utils.plotOne(0, rg, figname="pathGenInitial.png", isSequence=False)
    rg.decollideGrid()
    assert rg.getNCollisions() == 0
    if plot:
        utils.plotOne(0, rg, figname="pathGenDecollided.png", isSequence=False)
    rg.pathGenGreedy()
    # print("rg alg type", rg.algType, type(rg.algType), str(rg.algType))
    # sd = rg.robotGridSummaryDict()
    # for d in sd["robotDict"].values():
    #     print(d["id"], d["alpha"], d["beta"])
    # print("deadlocks", rg.deadlockedRobots())
    assert not rg.didFail
    rg.smoothPaths(smoothPts)
    rg.simplifyPaths()
    # rg.verifySmoothed()
    # assert rg.smoothCollisions > 100
    # print(rg.smoothCollisions)
    rg.shrinkCollisionBuffer(collisionShrink)
    # rg.verifySmoothed()
    # assert rg.smoothCollisions == 0
    print(rg.smoothCollisions)
    if plot:
        utils.plotOne(0, rg, figname="donepaht.png", isSequence=False)
        utils.plotPaths(rg, downsample=downsample, filename="pathGen.mp4")



def test_withDefulatArgs(plot=False):
    rg = RobotGridNominal() # this is the test, that no args still works

    for rID in rg.robotDict:
        robot = rg.getRobot(rID)
        robot.setXYUniform()
        robot.setDestinationAlphaBeta(0, 180)
    assert rg.getNCollisions() > 10
    rg.decollideGrid()
    assert rg.getNCollisions() == 0
    rg.pathGenGreedy()
    if plot:
        utils.plotPaths(rg, filename="test_default.mp4")


if __name__ == "__main__":
    # pytest won't run these, run by hand for the plot output
    # test_hexDeadlockedPath(plot=True)
    # test_pathGen(plot=True)
    # test_filledHexDeadlockedPath(plot=True)
    # test_filledHexPath(plot=True)
    # test_pathGen(plot=True)
    test_pathGen(plot=True)





import matplotlib
import matplotlib.pyplot as plt
import numpy
from shapely.geometry import LineString, Point
from descartes import PolygonPatch
from multiprocessing import Pool, cpu_count
from subprocess import Popen
import glob
import os
from .cKaiju import RobotGrid
import coordio

matplotlib.use('Agg')

# internalBuffer = 1.5
rg = None # need global because C++ obj can't be pickled for multiprocessing

def plotTraj(r, figprefix="traj_", dpi=500):
    # r is a robot
    spa = numpy.array(r.smoothedAlphaPath)
    spb = numpy.array(r.smoothedBetaPath)
    rpa = numpy.array(r.alphaPath)
    rpb = numpy.array(r.betaPath)
    aRDP = numpy.array(r.simplifiedAlphaPath);
    bRDP = numpy.array(r.simplifiedBetaPath);


    av = numpy.array(r.alphaVel)
    bv = numpy.array(r.betaVel)
    vSteps = numpy.arange(len(av))
    sav = numpy.array(r.smoothAlphaVel)
    sbv = numpy.array(r.smoothBetaVel)
    ss = numpy.arange(len(sav))

    # print("plotting", r.id)
    # print("alpha start", rpa[0,:] - aRDP[0,:])
    # print("alpha end", rpa[-1,:] - aRDP[-1,:])
    # print("beta start", rpb[0,:] - bRDP[0,:])
    # print("beta end", rpb[-1,:] - bRDP[-1,:])

    fig, ax = plt.subplots(2,1, figsize=(10,10))


    ax[0].plot(rpa[:,0], rpa[:,1], linewidth=0.2, label="rough alpha", alpha=0.8)
    ax[0].plot(rpb[:,0], rpb[:,1], linewidth=0.2, label="rough beta", alpha=0.8)
    ax[0].plot(spa[:,0], spa[:,1], 'k-', linewidth=0.2, label="smooth alpha")
    ax[0].plot(spb[:,0], spb[:,1], 'k-', linewidth=0.2, label="smooth beta")
    ax[0].plot(
        aRDP[:,0], aRDP[:,1], 'oc-', linewidth=0.2, markeredgewidth=0.4,
        fillstyle="none", markersize=2, label="RDP alpha", alpha=0.7
    )
    ax[0].plot(
        bRDP[:,0], bRDP[:,1], 'oc-', linewidth=0.2, markeredgewidth=0.4,
        fillstyle="none", markersize=2, label="RDP beta", alpha=0.7
    )
    ax[0].legend()

    ax[1].plot(vSteps, av, linewidth=0.2, label="alphaVel", alpha=0.4)
    ax[1].plot(vSteps, bv, linewidth=0.2, label="betaVel", alpha=0.4)
    ax[1].plot(ss, sav, 'k-', linewidth=0.2, label="smoothAlpha")
    ax[1].plot(ss, sbv, 'k-', linewidth=0.2, label="smoothBeta")

    ax[1].legend()
    # plt.legend()


    plt.savefig(figprefix+"robot_%s.pdf"%r.id, dpi=dpi)
    plt.close()


def plotOne(
    step, robotGrid=None, figname=None, isSequence=True,
    plotTargets=False, plotRobotIDs=False, dpi=100,
    xlim=None, ylim=None, highlightRobot=None, returnax=False
):
    global rg
    if hasattr(step, "__len__"):
        fig = step[1]
        step = step[0]
    else:
        fig = step

    if robotGrid is not None:
        rg = robotGrid
    plt.figure(figsize=(10,10))
    ax = plt.gca()
    ax.set_facecolor('lightgray')
    maxX = 0
    maxY = 0
    for robotID, robot in rg.robotDict.items():
        # account for xy calibration offsets in base
        # pos of robot in wok space
        hatch = None
        wokBase = coordio.libcoordio.tangentToWok(
            [0,0,0],
            robot.basePos,
            robot.iHat,
            robot.jHat,
            robot.kHat,
            robot.elementHeight,
            robot.scaleFac,
            robot.dxyz[0],
            robot.dxyz[1],
            robot.dxyz[2]
        )

        basePos = wokBase[:-1]
        # basePos = robot.basePos

        if basePos[0] > maxX:
            maxX = basePos[0]
        if basePos[1] > maxY:
            maxY = basePos[1]
        if isSequence:
            alphaX = robot.roughAlphaX[step][1]
            alphaY = robot.roughAlphaY[step][1]
            betaX = robot.roughBetaX[step][1]
            betaY = robot.roughBetaY[step][1]
            onTarget = False
            # onTarget = robot.onTargetVec[step]
        else:
            # step input is ignored!!
            alphaPoint = robot.collisionSegWokXYZ[0]
            betaPoint = robot.collisionSegWokXYZ[1]
            alphaX = alphaPoint[0]
            alphaY = alphaPoint[1]
            betaX = betaPoint[0]
            betaY = betaPoint[1]
            onTarget = robot.score() == 0
        plt.plot(
            [basePos[0], alphaX], [basePos[1], alphaY],
            color='black', linewidth=2, alpha=0.5
        )

        topCollideLine = LineString(
            [(alphaX, alphaY), (betaX, betaY)]
        ).buffer(robot.collisionBuffer, cap_style=1)
        topcolor = 'blue'
        edgecolor = 'black'
        if not robot.isAssigned():
            topcolor = "skyblue"
        if onTarget:
            topcolor = "gold"
        if not isSequence and rg.isCollided(robotID):
            # collision trumps not assigned
            topcolor = "red"
        if highlightRobot is not None:
            if highlightRobot == robotID:
                topcolor = "violet"
                # edgecolor = "red"
                hatch=None
            elif hasattr(highlightRobot, "__len__") and robotID in highlightRobot:
                topcolor = "violet"
                # edgecolor = "red"
                hatch = None
        if robot.isOffline:
            topcolor = "black"
        patch = PolygonPatch(
            topCollideLine, fc=topcolor, ec=edgecolor, hatch=hatch, alpha=0.5, zorder=10
        )
        ax.add_patch(patch)
        if plotRobotIDs:
            strID = "P" + ("%i"%robot.id).zfill(4)
            xm = (alphaX + betaX)/2
            ym = (alphaY + betaY)/2
            dx = alphaX-betaX
            dy = alphaY-betaY
            rot = numpy.degrees(numpy.arctan2(dy, dx))
            if alphaX < betaX:
                rot += 180
            props = {"ha": "center", "va": "center", "fontsize": 2}
            ax.text(xm, ym, strID, props, rotation=rot)

    for fiducialID, fiducial in rg.fiducialDict.items():

        fPoint = Point(
            fiducial.xyzWok[0], fiducial.xyzWok[1]
        ).buffer(fiducial.collisionBuffer) #, cap_style=1)

        patch = PolygonPatch(fPoint, fc="aquamarine", ec="black", alpha=0.8, zorder=10)
        ax.add_patch(patch)
    for gfa in rg.gfaDict.values():

        gLine = LineString(
            [gfa.collisionSegWokXYZ[0], gfa.collisionSegWokXYZ[1]]
        ).buffer(gfa.collisionBuffer, cap_style=1)
        patch = PolygonPatch(gLine, fc="aquamarine", ec="black", alpha=0.5, zorder=10)
        ax.add_patch(patch)

    if xlim is not None:
        ax.set_xlim(xlim)
    if ylim is not None:
        ax.set_ylim(ylim)
    ax.set_xlim([-maxX-50, maxX+50])
    ax.set_ylim([-maxY-50, maxY+50])
    ax.set_aspect("equal")

    if returnax:
        return ax

    if figname is None:
        figname = "step_%04d.png"%(fig)
    plt.savefig(figname, dpi=dpi)
    plt.close()

def plotPaths(robotGrid, downsample=None, filename=None):
    # figure out how to downsample paths
    global rg
    rg = robotGrid
    steps = list(range(robotGrid.nSteps)) # not sure why rg.nSteps is broken
    if downsample is not None:
        steps = steps[::downsample] #+ [steps[-1]]
        # throwout last step, sometimes is out of range index depending
        # on which downsample was selected
        steps = steps[:-1]


    figs = list(range(len(steps)))
    stepfigs = list(zip(steps, figs))
    print("plotting steps", figs[-1])
    # down sample if nframes specified

    for stepfig in stepfigs:
        plotOne(stepfig, robotGrid=rg)

    # p = Pool(cpu_count())
    # p.map(plotOne, stepfigs)

    # for stepFig in stepfigs:
    #     print("stepfig", stepFig)
    #     plotOne(stepFig)

    fps = 30 # frames per second
    if filename is None:
        filename = "example.mp4"
    # clobber file if it already exists
    if os.path.exists(filename):
        os.remove(filename)
    args = ['ffmpeg', '-r', '%i'%fps, '-f', 'image2', '-i', 'step_%04d.png',
            '-pix_fmt', 'yuv420p', filename]

    movie = Popen(args)
    movie.wait()

    # clean up imgs
    imgs = glob.glob("step*.png")
    for img in imgs:
        os.remove(img)


def hexFromDia(nDia, pitch=22.4, rotAngle=0):
    """
    inputs:
    nDia: is points along the equator of the hex, must be odd

    returns:
    xArr, yArr: both nd arrays of x and y coordinates for each
    position in the hex. Units are mm and the center of hex is 0,0
    """
    # first determine the xy positions for the center line
    nRad = int((nDia + 1) / 2)
    xPosDia = numpy.arange(-nRad + 1, nRad) * pitch
    yPosDia = numpy.zeros(nDia)

    vertShift = numpy.sin(numpy.radians(60.0)) * pitch
    horizShift = numpy.cos(numpy.radians(60.0)) * pitch
    # populate the top half of the hex
    # determine number of rows above the diameter row
    xUpperHalf = []
    yUpperHalf = []
    for row in range(1, nRad):
        xPos = xPosDia[:-1 * row] + row * horizShift
        yPos = yPosDia[:-1 * row] + row * vertShift
        xUpperHalf.extend(xPos)
        yUpperHalf.extend(yPos)
    xUpperHalf = numpy.asarray(xUpperHalf).flatten()
    yUpperHalf = numpy.asarray(yUpperHalf).flatten()

    xLowerHalf = xUpperHalf[:]
    yLowerHalf = -1 * yUpperHalf[:]

    xAll = numpy.hstack((xLowerHalf, xPosDia, xUpperHalf)).flatten()
    yAll = numpy.hstack((yLowerHalf, yPosDia, yUpperHalf)).flatten()

    # rotate grid?
    if rotAngle:
        cos = numpy.cos(numpy.radians(rotAngle))
        sin = numpy.sin(numpy.radians(rotAngle))
        rotMat = numpy.array([
            [cos, sin],
            [-sin, cos]
        ])
        xyAll = numpy.array([xAll, yAll]).T
        xyRot = numpy.dot(xyAll, rotMat)
        xAll = xyRot[:,0]
        yAll = xyRot[:,1]
    # return arrays in a sorted order, rasterized. 1st position is lower left
    # last positon is top right
    ind = numpy.lexsort((xAll, yAll))
    return xAll[ind], yAll[ind]














import pytest

import kaiju
from kaiju.robotGrid import RobotGridNominal
import coordio

fpZ = coordio.defaults.POSITIONER_HEIGHT

def test_filledHex():
    # should make a grid
    rg = RobotGridNominal()


def test_robot_array():
    # should make an array from robots and convert back the same
    rg = RobotGridNominal()
    rarray = rg.robot_array()
    rg2 = RobotGridNominal()
    rg2.robot_fromarray(rarray)


def test_add_targets():
    # should add targets OK
    rg = RobotGridNominal()
    xyzWok1 = [10, 10, fpZ]
    xyzWok2 = [30, 30, fpZ]
    rg.addTarget(targetID=1, xyzWok=xyzWok1, priority=1.,
                 fiberType=kaiju.cKaiju.ApogeeFiber)
    rg.addTarget(targetID=2, xyzWok=xyzWok2, priority=1.,
                 fiberType=kaiju.cKaiju.BossFiber)


def test_target_array():
    # should add targets OK
    rg = RobotGridNominal()
    xyzWok1 = [10, 10, fpZ]
    xyzWok2 = [30, 30, fpZ]
    rg.addTarget(targetID=1, xyzWok=xyzWok1, priority=1.,
                 fiberType=kaiju.cKaiju.ApogeeFiber)
    rg.addTarget(targetID=2,xyzWok=xyzWok2, priority=1.,
                 fiberType=kaiju.cKaiju.BossFiber)
    ts = rg.target_array()
    rg2 = RobotGridNominal()
    rg2.target_fromarray(ts)
    for id in [1, 2]:
        assert rg2.targetDict[id].xWok == rg.targetDict[id].xWok
        assert rg2.targetDict[id].yWok == rg.targetDict[id].yWok
        assert rg2.targetDict[id].priority == rg.targetDict[id].priority
        assert rg2.targetDict[id].fiberType == rg.targetDict[id].fiberType
        assert rg2.targetDict[id].id == rg.targetDict[id].id



from kaiju.robotGrid import RobotGridCalib
from kaiju import utils
import numpy

angStep = 0.05
collisionBuffer = 2.3
epsilon = 2.2*angStep
hasApogee = True


def test_offlineRobots(plot=False):
    # xPos, yPos = utils.hexFromDia(35, pitch=22.4)
    seed = 6
    rg = RobotGridCalib(
        stepSize=angStep,
        epsilon=epsilon, seed=seed
    )
    rg.setCollisionBuffer(collisionBuffer)


    for rID in rg.robotDict:
        robot = rg.getRobot(rID)
        if rID == 1255:
            robot.setAlphaBeta(305, 238.29)
            robot.isOffline = True
        elif rID == 717:
            robot.setAlphaBeta(0,180)
            robot.isOffline = True
        elif rID == 1367:
            robot.setAlphaBeta(0, 164.88)
            robot.isOffline = True
        elif rID == 398:
            robot.setAlphaBeta(0, 152.45)
            robot.isOffline = True
        else:
            robot.setXYUniform()

    rg.decollideGrid()

    # for robot in rg.robotDict.values():
    #     if numpy.random.uniform() > 0.998:

    #         robot.isOffline = True

    for robot in rg.robotDict.values():
        robot.setDestinationAlphaBeta(0, 180)

    assert rg.getNCollisions() == 0
    rg.pathGenMDP(0.8, 0.2)
    print("didfai", rg.didFail)
    if plot:
        utils.plotPaths(rg, downsample=int(numpy.floor(10 / angStep)), filename="offline.mp4")

if __name__ == "__main__":
    test_offlineRobots(plot=True)


from kaiju import RobotGridCalib
from kaiju.utils import plotOne
import numpy


def test_gfaCollision(plot=False):
    rg = RobotGridCalib()

    for alphaAng in numpy.linspace(0, 360, 100): #[0, 60, 120, 180, 240, 300, 360]:
        for r in rg.robotDict.values():
            r.setAlphaBeta(alphaAng, 0)
        # assert rg.getNCollisions() == 6
        if plot:
            plotOne(0, rg, "gfa%i_collide.png"%alphaAng, False)
        rg.decollideGrid()
        assert rg.getNCollisions() == 0
        if plot:
            plotOne(0, rg, "gfa%i_decollide.png"%alphaAng, False)


if __name__ == "__main__":
    test_gfaCollision(plot=True)


import os
import sys
import urllib.request

from setuptools import Extension, find_packages, setup
from setuptools.command.build_ext import build_ext as _build_ext


COORDIO_BASE = 'https://raw.githubusercontent.com/sdss/coordio/master/'

class getPybindInclude(object):
    """Helper class to determine the pybind11 include path
    The purpose of this class is to postpone importing pybind11
    until it is actually installed, so that the ``get_include()``
    method can be invoked.
    https://github.com/pybind/python_example/blob/master/setup.py
    """

    def __init__(self, user=False):
        self.user = user

    def __str__(self):
        import pybind11
        return pybind11.get_include(self.user)


def getIncludes():
    return [
        'python/kaiju/include',
        getPybindInclude(),
        getPybindInclude(user=True),
    ]


def getVersion():
    with open("python/kaiju/__version__.py", "r") as f:
        lines = f.readlines()
    for line in lines:
        if line.startswith("__version__"):
            v = line.split("=")[-1].strip().strip('"').strip("'")
            return v


def getSources():
    return [
        'src/cKaiju.cpp',
        'src/robot.cpp',
        'src/robotGrid.cpp',
        'src/utils.cpp',
        'src/target.cpp',
        'src/fiducial.cpp',
        'src/gfa.cpp',
        'src/conv.cpp'
    ]


extra_compile_args = ["--std=c++11", "-fPIC", "-v", "-O3"]
extra_link_args = []
if sys.platform == 'darwin':
    extra_compile_args += ['-stdlib=libc++', '-mmacosx-version-min=10.9']
    extra_link_args = ["-v", '-mmacosx-version-min=10.9', '-L.']

    from distutils import sysconfig
    vars = sysconfig.get_config_vars()
    vars['LDSHARED'] = vars['LDSHARED'].replace('-bundle', '-dynamiclib')

extensions = [
    Extension(
        'kaiju.cKaiju',
        include_dirs=getIncludes(),
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
        sources=getSources())
]


# cKaiju relies on some funcitons defined in libcoordio in coordio. Since linking against
# that library directly is painful, we just download the latest files from GitHub
# and add them to cKaiju ...
class build_ext(_build_ext):

    def run(self):

        DIRNAME = os.path.dirname(__file__)

        with urllib.request.urlopen(COORDIO_BASE + 'src/coordio/include/coordio.h') as r:
            data = r.read()
            with open(os.path.join(DIRNAME, 'python/kaiju/include/coordio.h'), 'wb') as f:
                f.write(data)

        with urllib.request.urlopen(COORDIO_BASE + 'cextern/conv.cpp') as r:
            data = r.read()
            with open(os.path.join(DIRNAME, 'src/conv.cpp'), 'wb') as f:
                f.write(data)

        super().run()


def runSetup(packages, requirements):
    setup(
        name="sdss-kaiju",
        version=getVersion(),
        license="BSD3",
        author="Conor Sayres",
        author_email="csayres@uw.edu",
        description="Collision Avoidance for SDSS-V Positioners",
        long_description=open('README.rst').read(),
        packages=packages,
        cmdclass={'build_ext': build_ext},
        package_dir={'': 'python'},
        package_data={'kaiju': ['python/kaiju/include/*']},
        include_package_data=True,
        url="https://github.com/sdss/kaiju",
        keywords="astronomy software",
        ext_modules=extensions,
        install_requires=requirements,
        setup_requires=[
            "sdss-coordio>=1.2.1",
            "pybind11>=2.2.4,<3"
        ],
        classifiers=[
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: BSD License',
            'Natural Language :: English',
            'Operating System :: OS Independent',
            'Programming Language :: Python',
        ]
    )


requirementsFile = os.path.join(os.path.dirname(__file__), "requirements.txt")
requirements = [line.strip() for line in open(requirementsFile)]
packages = find_packages(where="python")
print("found packages", packages)
runSetup(packages, requirements)



import pytest
import numpy
from kaiju import RobotGridCalib
from kaiju import utils
import time
numpy.random.seed(0)


def test_explode(plot=False):

    rg = RobotGridCalib(seed=350)

    rg.setCollisionBuffer(2.5)
    for robot in rg.robotDict.values():
        robot.setXYUniform()
        robot.setDestinationAlphaBeta(0, 180)


    rg.decollideGrid()

    assert rg.getNCollisions() == 0
    tstart = time.time()
    rg.pathGenExplode(100)
    print("pathgen took", time.time()-tstart)

    if plot:
        utils.plotPaths(rg, filename="explode.mp4")


if __name__ == "__main__":
    test_explode(plot=True)



import pytest
import matplotlib.pyplot as plt
import numpy
from kaiju import RobotGrid
from kaiju import utils
import time
import json
import pickle

nDia = 15
angStep = 3
collisionBuffer = 2
epsilon = angStep * 2
hasApogee = True


def test_forwardGreedy(plot=False):
    xPos, yPos = utils.hexFromDia(25, pitch=22.4)
    seed = 1
    rg = RobotGrid(
        stepSize=angStep,
        epsilon=epsilon, seed=seed
    )

    for robotID, (x, y) in enumerate(zip(xPos, yPos)):
        rg.addRobot(robotID, str(robotID), [x, y, 0], hasApogee)
    rg.setCollisionBuffer(collisionBuffer)
    rg.initGrid()
    for rID in rg.robotDict:
        robot = rg.getRobot(rID)
        robot.setXYUniform()
    assert rg.getNCollisions() > 10

    rg.decollideGrid()

    for robot in rg.robotDict.values():
        robot.setDestinationAlphaBeta(robot.alpha, robot.beta)
        robot.setAlphaBeta(0, 180)
    assert rg.getNCollisions() == 0
    rg.pathGenGreedy()
    if plot:
        utils.plotPaths(rg, filename="forwardGreedy.mp4")

def test_reverseGreedy(plot=False):
    xPos, yPos = utils.hexFromDia(25, pitch=22.4)
    seed = 1
    rg = RobotGrid(
        stepSize=angStep,
        epsilon=epsilon, seed=seed
    )

    for robotID, (x, y) in enumerate(zip(xPos, yPos)):
        rg.addRobot(robotID, str(robotID), [x, y, 0], hasApogee)
    rg.setCollisionBuffer(collisionBuffer)
    rg.initGrid()
    for rID in rg.robotDict:
        robot = rg.getRobot(rID)
        robot.setXYUniform()
    assert rg.getNCollisions() > 10

    rg.decollideGrid()

    for robot in rg.robotDict.values():
        robot.setDestinationAlphaBeta(0, 180)
    assert rg.getNCollisions() == 0
    rg.pathGenGreedy()
    if plot:
        utils.plotPaths(rg, filename="reverseGreedy.mp4")

def test_forwardMDP(plot=False):
    xPos, yPos = utils.hexFromDia(25, pitch=22.4)
    seed = 1
    rg = RobotGrid(
        stepSize=angStep,
        epsilon=epsilon, seed=seed
    )

    for robotID, (x, y) in enumerate(zip(xPos, yPos)):
        rg.addRobot(robotID, str(robotID), [x, y, 0], hasApogee)
    rg.setCollisionBuffer(collisionBuffer)
    rg.initGrid()
    for rID in rg.robotDict:
        robot = rg.getRobot(rID)
        robot.setXYUniform()
    assert rg.getNCollisions() > 10

    rg.decollideGrid()
    print("N collisions 1", rg.getNCollisions())

    for robot in rg.robotDict.values():
        robot.setDestinationAlphaBeta(robot.alpha, robot.beta)
        robot.setAlphaBeta(0, 180)
    print("N collisions 2", rg.getNCollisions())
    # assert rg.getNCollisions() == 0
    rg.pathGenMDP(0.9, 0.1)
    if plot:
        utils.plotPaths(rg, filename="forwardMDP.mp4")

def test_reverseMDP(plot=False):
    greed = 0.9
    phobia = 0.1
    angStep = 1
    downsample = int(numpy.floor(3 / angStep))
    xPos, yPos = utils.hexFromDia(25, pitch=22.4)
    seed = 1
    rg = RobotGrid(
        stepSize=angStep,
        epsilon=epsilon, seed=seed
    )

    for robotID, (x, y) in enumerate(zip(xPos, yPos)):
        rg.addRobot(robotID, str(robotID), [x, y, 0], hasApogee)
    rg.setCollisionBuffer(collisionBuffer)
    rg.initGrid()
    for rID in rg.robotDict:
        robot = rg.getRobot(rID)
        robot.setXYUniform()
    assert rg.getNCollisions() > 10

    rg.decollideGrid()

    for robot in rg.robotDict.values():
        robot.setDestinationAlphaBeta(0, 180)
    assert rg.getNCollisions() == 0
    rg.pathGenMDP(greed, phobia)

    if plot:
        utils.plotPaths(rg, downsample=downsample, filename="reverseMDP.mp4")

def test_reverseSmoothMDP(plot=False):
    greed = 0.8
    phobia = 0.2
    downsample = int(numpy.floor(150 / angStep))
    xPos, yPos = utils.hexFromDia(17, pitch=22.4)
    seed = 1
    cb = 2.5
    cs = 0.04
    step = 0.1          # degrees per step in kaiju's rough path
    smoothPts = 10           # width of velocity smoothing window
    eps = step * 2.2
    rg = RobotGrid(
        stepSize=step,
        epsilon=eps, seed=seed
    )

    for robotID, (x, y) in enumerate(zip(xPos, yPos)):
        rg.addRobot(robotID, str(robotID), [x, y, 0], hasApogee, collisionBuffer=cb)
    rg.initGrid()
    for rID in rg.robotDict:
        robot = rg.getRobot(rID)
        robot.setXYUniform()
    assert rg.getNCollisions() > 10

    rg.decollideGrid()

    for robot in rg.robotDict.values():
        robot.setDestinationAlphaBeta(0, 180)
    assert rg.getNCollisions() == 0
    tstart = time.time()
    rg.pathGenMDP(greed, phobia)
    print("pathgen took", time.time()-tstart)
    rg.smoothPaths(smoothPts)
    rg.simplifyPaths()
    rg.shrinkCollisionBuffer(cs)
    # rg.verifySmoothed()


    assert rg.smoothCollisions == 0
    print("n smooth collisions", rg.smoothCollisions)

    if plot:
        for r in rg.robotDict.values():
            utils.plotTraj(r, "reverseSmoothMDP", dpi=250)
        utils.plotPaths(rg, downsample=downsample, filename="reverseSmoothMDP.mp4")


def test_setMDP(plot=False):

    greed = 0.8
    phobia = 0.2
    xPos, yPos = utils.hexFromDia(45, pitch=22.4)
    print("using ", len(xPos), "robots")
    # collisionBuffer = 3
    angStep = 0.5
    collisionBuffer = 3
    downsample = int(numpy.floor(3 / angStep))
    for seed in range(100):
        rg = RobotGrid(
            stepSize=angStep,
            epsilon=epsilon, seed=seed
        )

        for robotID, (x, y) in enumerate(zip(xPos, yPos)):
            rg.addRobot(robotID, str(robotID), [x, y, 0], hasApogee)
        rg.setCollisionBuffer(collisionBuffer)
        rg.initGrid()
        for rID in rg.robotDict:
            robot = rg.getRobot(rID)
            robot.setXYUniform()
        assert rg.getNCollisions() > 10

        rg.decollideGrid()

        for robot in rg.robotDict.values():
            robot.setDestinationAlphaBeta(90, 180)
        assert rg.getNCollisions() == 0
        rg.pathGenMDP(greed, phobia)

        deadlockedRobots = []
        for r in rg.robotDict.values():
            # if not r.onTargetVec[-1]:
            if r.score() > 0:
                deadlockedRobots.append(r.id)
        if len(deadlockedRobots):
            print("seed", seed, "failed with these", deadlockedRobots, "in ", rg.nSteps)
            break
        else:
            print("seed", seed, "didn't fail", rg.nSteps, " taken to solve")

    if plot:
        utils.plotPaths(rg, downsample=downsample, filename="reverseSetMDP.mp4")

def test_initialConfigs(plot=False):

    xPos, yPos = utils.hexFromDia(21, pitch=22.4)
    angStep = 0.1
    greed = 0.8
    phobia = 0.2
    downsample = int(numpy.floor(10 / angStep))
    rg = RobotGrid(
        stepSize=angStep,
        epsilon=epsilon, seed=1
    )
    for robotID, (x, y) in enumerate(zip(xPos, yPos)):
        rg.addRobot(robotID, str(robotID), [x, y, 0], hasApogee)
    rg.setCollisionBuffer(collisionBuffer)
    rg.initGrid()
    for rID in rg.robotDict:
        robot = rg.getRobot(rID)
        robot.setXYUniform()
    assert rg.getNCollisions() > 10
    if plot:
        utils.plotOne(-1, rg, figname="angStepO.png", isSequence=False)

    rg.decollideGrid()
    for robot in rg.robotDict.values():
        robot.setDestinationAlphaBeta(0, 180)
    if plot:
        utils.plotOne(-1, rg, figname="angStepD.png", isSequence=False)
    rg.pathGenMDP(greed, phobia)
    if plot:
        utils.plotOne(-1, rg, figname="angStepE.png", isSequence=False)
    if plot:
        utils.plotPaths(rg, downsample=downsample, filename="init.mp4")

def test_tofile(plot=False):
    xPos, yPos = utils.hexFromDia(37, pitch=22.4, rotAngle=90)
    print("n robots", len(xPos))
    angStep = 1
    greed = 0.8
    phobia = 0.2
    downsample = int(numpy.floor(3 / angStep))
    rg = RobotGrid(
        stepSize=angStep,
        epsilon=epsilon, seed=1
    )

    for robotID, (x, y) in enumerate(zip(xPos, yPos)):
        rg.addRobot(robotID, str(robotID), [x, y, 0], hasApogee)
    rg.setCollisionBuffer(collisionBuffer)
    rg.initGrid()
    for rID in rg.robotDict:
        robot = rg.getRobot(rID)
        robot.setXYUniform()
    assert rg.getNCollisions() > 10


    rg.decollideGrid()
    for robot in rg.robotDict.values():
        robot.setDestinationAlphaBeta(10, 170)
    rg.pathGenMDP(greed, phobia)
    # rg.smoothPaths(3)
    # rg.simplifyPaths()
    # rg.verifySmoothed()
    # t1 = time.time()
    # rg.summaryJSON("json.txt")
    # print("json took", (time.time()-t1))

    # t1 = time.time()
    # rg.summaryPickle("rg.pkl")
    # print("pickle took", (time.time()-t1))

    # t1 = time.time()
    # g = json.load(open("json.txt", "r"))
    # print("json load took", (time.time()-t1))

    # t1 = time.time()
    # g = pickle.load(open("rg.pkl", "rb"))
    # print("pickle load took", (time.time()-t1))
    if plot:
        utils.plotOne(-1, rg, figname="tofile.png", isSequence=False)
    # if plot:
    #     utils.plotPaths(rg, downsample=downsample, filename="tofile.mp4")


def test_fatty(plot=False):
    xPos, yPos = utils.hexFromDia(11, pitch=22.4)
    print("n robots", len(xPos))
    angStep = 0.1
    greed = 0.8
    phobia = 0.2
    collisionBuffer = 4
    downsample = int(numpy.floor(3 / angStep))
    rg = RobotGrid(
        stepSize=angStep,
        epsilon=epsilon, seed=1
    )

    for robotID, (x, y) in enumerate(zip(xPos, yPos)):
        rg.addRobot(robotID, str(robotID), [x, y, 0], hasApogee)
    rg.setCollisionBuffer(collisionBuffer)
    rg.initGrid()
    for rID in rg.robotDict:
        robot = rg.getRobot(rID)
        robot.setXYUniform()
        robot.setDestinationAlphaBeta(90,180)
    # assert rg.getNCollisions() > 10

    rg.decollideGrid()

    rg.pathGenMDP(greed, phobia)

    # rg.smoothPaths(3)
    # rg.simplifyPaths()
    # rg.verifySmoothed()
    if plot:
        utils.plotPaths(rg, downsample=downsample, filename="fatty.mp4")

if __name__ == "__main__":
    # test_forwardGreedy(plot=True)
    test_reverseSmoothMDP(plot=True)
    # test_reverseGreedy(plot=True)
    # test_forwardMDP(plot=True)
    # test_initialConfigs(plot=True)
    # test_initialConfigs(plot=True)

    # test_setMDP(plot=True)
    # test_tofile(plot=True)
    # test_forwardPathGen(plot=True)
    # test_reversePathGen(plot=True)
    # test_reversePathGen2(plot=True)
    # test_reversePathGen3(plot=True)
    # test_forwardPathGen2(plot=True)


